/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_TRANSACTION_NUMBER_H_
#define DRN_ACCOUNTING_TRANSACTION_NUMBER_H_

#include <QtCore>

#include <iosfwd>

#include <accounting_export.h>

class QDebug;
class QString;

namespace drn::accounting
{

/**
 * @brief The unique number that identifies a transaction in the accounting system.
 */
class DRN_ACCOUNTING_EXPORT TransactionNumber
{
	quint32 value_;

public:
	/**
	 * @brief Default initalises the object to an invalid numerical value.
	 */
	TransactionNumber();
	/**
	 * @brief Fully initalises the object with the supplied values.
	 *
	 * @exception TransactionError When the supplied new value is zero.
	 *
	 * @param value The number for the transaction.
	 */
	explicit TransactionNumber(const quint32& value);
	/**
	 * @brief Equality operator compares all members.
	 *
	 * @param other The other object whose members should all be equal.
	 *
	 * @return When all member variables are the same it'll give true, else false.
	 */
	bool operator==(const TransactionNumber& other) const;
	/**
	 * @brief Equality operator compares the instance against a value.
	 *
	 * @param other The numerical value that should be equal.
	 *
	 * @return When the member variable is the same it'll give true, else false.
	 */
	bool operator==(const quint32 other) const;
	/**
	 * @brief Inequality operator compares all members.
	 *
	 * @param other The other object whose members should be not equal.
	 *
	 * @return When any member variables is not the same it'll give true, else false.
	 */
	bool operator!=(const TransactionNumber& other) const;
	/**
	 * @brief Inequality operator compares the instance against a value.
	 *
	 * @param other The numerical value that should not be equal.
	 *
	 * @return When the member variable is not the same it'll give true, else false.
	 */
	bool operator!=(const quint32 other) const;
	/**
	 * @brief Less-than comparison between the current instance member variables and another.
	 *
	 * @param other The other object whose members should all be less-than.
	 *
	 * @return When all member variables in the current instance is less than the supplied.
	 */
	bool operator<(const TransactionNumber& other) const;
	/**
	 * @brief Greater-than comparison between the current instance member variables and another.
	 *
	 * @param other The other object whose members should all be greater-than.
	 *
	 * @return When all member variables in the current instance is greater than the supplied.
	 */
	bool operator>(const TransactionNumber& other) const;
	/**
	 * @brief Increment the object by one.
	 *
	 * @return The instance after the increment operation is performed.
	 */
	TransactionNumber& operator++();
	/**
	 * @brief Increment the object by one.
	 *
	 * @return The instance before the increment operation is performed.
	 */
	TransactionNumber operator++(int);
	/**
	 * @brief The current weak type numerical value of the transaction.
	 */
	const quint32& integer() const;

	/// The number that indicates that it is invalid and empty.
	static const quint32 invalid_;
};

DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const TransactionNumber& tn);
DRN_ACCOUNTING_EXPORT ::QDebug operator<<(::QDebug o, const TransactionNumber& tn);
DRN_ACCOUNTING_EXPORT ::QString presentationText(const TransactionNumber& tn);

}

#endif
