/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_SURVEYING_COMPLETED_SURVEY_H_
#define DRN_SURVEYING_COMPLETED_SURVEY_H_

#include <pecunia/Money.h>

#include <QString>

#include <iosfwd>
#include <map>
#include <optional>

#include <budgeting/BudgetItemIdentifier.h>
#include <banking/BankAccount.h>
#include <banking/Reconciliation.h>

#include <surveying_export.h>


class QDate;
class QDebug;

namespace drn::accounting
{

class AccountCode;
class AccountNumber;

}
namespace drn::foundation
{

class Percentage;

}
namespace drn::surveying
{

class DistributedAmount;

class DRN_SURVEYING_EXPORT CompletedSurvey
{
	banking::ReconciledBankAccount reconciled_;
	std::map<budgeting::BudgetItemIdentifier, pecunia::currency::Money> distribution_;
	pecunia::currency::Money nonDistributed_;
	std::optional<::QString> checksum_;

public:
	CompletedSurvey() = default;
	/**
	 * @brief Partial initialisation constructor for creating a new survey.
	 *
	 * @param ba The bank account that is reconciled.
	 * @param currency The currency held within the bank account.
	 */
	CompletedSurvey(banking::BankAccount ba, const pecunia::currency::Currency& currency);
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception SurveyError When a supplied checksum value does not match the calculated value.
	 * @exception SurveyError When the reconciled bank account currency does not match the
	 * non-distributed currency.
	 *
	 * @param reconciled The bank account that was reconciled.
	 * @param distribution The distribution of the balance in the account.
	 * @param checksum The value that represents an object whose values were verified after having
	 * completed a survey. This is in lieu of verifying the balances upon object creation since
	 * exchange rates change over time preventing verification in the future.
	 * @param nonDistributed The funds of the balance not placed into the distribution.
	 */
	CompletedSurvey(
		banking::ReconciledBankAccount reconciled,
		std::map<
			budgeting::BudgetItemIdentifier,
			pecunia::currency::Money
		> distribution,
		pecunia::currency::Money nonDistributed,
		std::optional<::QString> checksum
	);
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception SurveyError When a supplied checksum value does not match the calculated value.
	 * @exception SurveyError When the reconciled bank account currency does not match the
	 * non-distributed currency.
	 *
	 * @param reconciled The bank account that was reconciled.
	 * @param distribution The distribution of the amounts in the account.
	 * @param checksum The value that represents an object whose values were verified after having
	 * completed a survey. This is in lieu of verifying the balances upon object creation since
	 * exchange rates change over time preventing verification in the future.
	 * @param nonDistributed The funds of the balance not placed into the distribution.
	 */
	CompletedSurvey(
		banking::ReconciledBankAccount reconciled,
		const std::map<budgeting::BudgetItemIdentifier, DistributedAmount>& distribution,
		pecunia::currency::Money nonDistributed,
		std::optional<::QString> checksum
	);
	CompletedSurvey(const CompletedSurvey& other) noexcept;
	CompletedSurvey(CompletedSurvey&& other) noexcept;
	~CompletedSurvey() noexcept;
	CompletedSurvey& operator=(const CompletedSurvey& other) noexcept;
	CompletedSurvey& operator=(CompletedSurvey&& other) noexcept;
	/// The reconciled bank account whose survey is completed.
	[[nodiscard]]
	const banking::ReconciledBankAccount& reconciled() const noexcept;
	/**
	 * @brief The distribution of unused funds (of the bank balance) in the bank account.
	 */
	[[nodiscard]]
	const std::map<
		budgeting::BudgetItemIdentifier,
		pecunia::currency::Money
	>& distribution() const noexcept;
	/**
	 * @brief The balance distribution specifically not distributed into the the distributable
	 * budget items.
	 */
	[[nodiscard]]
	const pecunia::currency::Money& nonDistributed() const noexcept;
	/**
	 * @brief The value that represents a valid object after having completed a survey.
	 */
	[[nodiscard]]
	const std::optional<::QString>& checksum() const noexcept;
	/**
	 * @brief Updates the survey with the completed values.
	 *
	 * @exception SurveyError When the reconciled bank balance does not match the amount of
	 * distributed unused funds combined with the non-distributed funds.
	 * @exception SurveyError When the bank's account currency is not the same as the
	 * non-distributed currency.
	 *
	 * @param completedOn The date the balance was confirmed.
	 * @param reconciledBalance The balance as reported in the bank account.
	 * @param nonDistributed The funds of the balance not placed into the distribution.
	 * @param distribution The distribution of unused funds (bank balance) in the bank account.
	 * @param balanceWindow The +/- percentage window in whtich to be considered balanced.
	 *
	 * @post Upon successful survey completion, the validity value is computed and stored in the
	 * object.
	 */
	void completed(
		const ::QDate& completedOn,
		const pecunia::currency::Money& reconciledBalance,
		const std::map<budgeting::BudgetItemIdentifier, DistributedAmount>& distribution,
		const pecunia::currency::Money& nonDistributed,
		const foundation::Percentage& balanceWindow
	);
	void renamed(const banking::BankName& original, const banking::BankName& newName);
	void renamed(
		const accounting::AccountNumber& original,
		const accounting::AccountNumber& changed
	);
	void renamed(const budgeting::BudgetItemIdentifier& id, const budgeting::BudgetSource& newName);
};

class DRN_SURVEYING_EXPORT CompletedSurveys final : std::map<banking::BankAccount, CompletedSurvey>
{
public:
	explicit CompletedSurveys(std::map<banking::BankAccount, CompletedSurvey> surveys);
	using std::map<banking::BankAccount, CompletedSurvey>::at;
	using std::map<banking::BankAccount, CompletedSurvey>::begin;
	using std::map<banking::BankAccount, CompletedSurvey>::cbegin;
	using std::map<banking::BankAccount, CompletedSurvey>::cend;
	using std::map<banking::BankAccount, CompletedSurvey>::const_iterator;
	using std::map<banking::BankAccount, CompletedSurvey>::count;
	using std::map<banking::BankAccount, CompletedSurvey>::empty;
	using std::map<banking::BankAccount, CompletedSurvey>::end;
	using std::map<banking::BankAccount, CompletedSurvey>::erase;
	using std::map<banking::BankAccount, CompletedSurvey>::map;
	using std::map<banking::BankAccount, CompletedSurvey>::size;
	[[nodiscard]]
	std::optional<CompletedSurvey> lookUpSurvey(const banking::BankAccount& ba) const;
	[[nodiscard]]
	std::optional<CompletedSurvey> lookUpSurvey(
		const banking::BankName& bn,
		const accounting::AccountNumber& an
	) const;
	bool upsert(CompletedSurvey survey);
};

DRN_SURVEYING_EXPORT bool operator<(
	const CompletedSurvey& lhs,
	const CompletedSurvey& rhs
);
DRN_SURVEYING_EXPORT bool operator==(
	const CompletedSurvey& lhs,
	const CompletedSurvey& rhs
);
DRN_SURVEYING_EXPORT bool operator!=(
	const CompletedSurvey& lhs,
	const CompletedSurvey& rhs
);
DRN_SURVEYING_EXPORT bool operator==(
	const CompletedSurveys& lhs,
	const CompletedSurveys& rhs
);
DRN_SURVEYING_EXPORT bool operator!=(
	const CompletedSurveys& lhs,
	const CompletedSurveys& rhs
);
DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& o, const CompletedSurvey& rs);
DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& o, const CompletedSurveys& rss);
DRN_SURVEYING_EXPORT ::QDebug operator<<(::QDebug out, const CompletedSurvey& rs);
DRN_SURVEYING_EXPORT ::QDebug operator<<(::QDebug out, const CompletedSurveys& rss);

}

#endif
